@@ -63,6 +63,9 @@ sub init_symbols {
     $parser->symbol('DEFAULT') ->set_std(\&std_set);
     $parser->symbol('CALL')    ->set_std(\&std_call);
 
+    $parser->symbol('NEXT')    ->set_std( $parser->can('std_next') );
+    $parser->symbol('LAST')    ->set_std( $parser->can('std_last') );
+
     $parser->symbol('MACRO') ->set_std(\&std_macro);
     $parser->symbol('BLOCK');
     $parser->symbol('WRAPPER')->set_std(\&std_wrapper);
@@ -73,7 +76,7 @@ sub init_symbols {
     # unsupported directives
     my $nos = $parser->can('not_supported');
     foreach my $keyword (qw(
-            INSERT PROCESS PERL RAWPERL TRY THROW NEXT LAST RETURN
+            INSERT PROCESS PERL RAWPERL TRY THROW RETURN
             STOP CLEAR META TAGS DEBUG VIEW)) {
         $parser->symbol($keyword)->set_std($nos);
     }
@@ -372,11 +375,7 @@ sub set_list {
 sub std_get {
     my($parser, $symbol) = @_;
 
-    my $stmt = $symbol->clone(
-        id    => 'print',
-        arity => 'command',
-        first => [ $parser->expression(0) ],
-    );
+    my $stmt = $parser->print( $parser->expression(0) );
     return $parser->finish_statement($stmt);
 }
 
@@ -552,14 +551,7 @@ sub std_filter {
         $filter = 'unmark_raw';
     }
     my $callfilter = $parser->call($filter, $callmacro);
-
-    my $print = $parser->symbol('print')->clone(
-        arity => 'command',
-        first => [$callfilter],
-        line  => $symbol->line,
-    );
-
-    return( $proc, $print );
+    return( $proc, $parser->print($callfilter) );
 }
 
 no Any::Moose;
@@ -650,7 +642,7 @@ Loop iterators are partially supported.
         [%- IF loop.is_first -%]
         <first>
         [%- END -%]
-        * [% loop.index %]
+        * [% loop.index %]  # 0 origin
         * [% loop.count     # loop.index + 1 %]
         * [% loop.body      # alias to arrayref %]
         * [% loop.size      # loop.body.size %]
